from .decorators import (
    PUT, POST, DELETE, annotate
)
from .schematics_serializer import SchematicsSerializer
from .contenttype_serializers import get_default_serializer_set
from .exceptions import APIException


class TransmuteContext(object):
    """
    TransmuteContext encapsulates all the configuration that transmute's
    core framework requires for all functionality it provides.

    When it's required to provide some customization on some
    of web-transmute's global values, a context can be used.

    In the case of multiple applications with
    """

    def __init__(self, serializers=None, contenttype_serializers=None):
        self._serializers = serializers or SchematicsSerializer()
        self._contenttype_serializers = (
            contenttype_serializers or
            get_default_serializer_set()
        )

    @property
    def serializers(self):
        return self._serializers

    @property
    def contenttype_serializers(self):
        return self._contenttype_serializers

# a global context is provided, if a singleton is sufficient
# or deviations from the defaults are unnescessary
default_context = TransmuteContext()



from collections import namedtuple


def get_signature(argspec):
    """
    retrieve a FunctionSignature object
    from the argspec and the annotations passed.
    """
    attributes = (getattr(argspec, "args", []) +
                  getattr(argspec, "keywords", []))
    defaults = argspec.defaults or []

    arguments, keywords = [], {}

    attribute_list = attributes[:-len(defaults)] if len(defaults) != 0 else attributes[:]
    for name in attribute_list:
        if name == "self":
            continue
        typ = argspec.annotations.get(name)
        arguments.append(Argument(name, NoDefault, typ))

    if len(defaults) != 0:
        for name, default in zip(attributes[-len(defaults):], defaults):
            typ = argspec.annotations.get(name)
            keywords[name] = Argument(name, default, typ)

    return FunctionSignature(arguments, keywords)


class NoDefault(object):

    def __str__(self):
        return "NoDefault"

    def __repr__(self):
        return "NoDefault"

NoDefault = NoDefault()


Argument = namedtuple("Argument", ["name", "default", "type"])


class FunctionSignature(object):

    NoDefault = NoDefault

    def __init__(self, args, kwargs):
        self.args = args
        self.kwargs = kwargs



import yaml
from .interface import ContentTypeSerializer


class YamlSerializer(ContentTypeSerializer):

    @staticmethod
    def to_type(data):
        """
        should return back a bytes (or string in python 2),
        representation of your object, to be used in e.g. response
        bodies.
        """
        return yaml.dump(data, default_flow_style=False).encode("UTF-8")

    @staticmethod
    def from_type(raw_bytes):
        """
        given a bytes object, should return a base python data
        structure that represents the object.
        """
        return yaml.load(raw_bytes)

    @staticmethod
    def can_handle(content_type_name):
        """
        given a content type, returns true if this serializer
        can convert bodies of the given type.
        """
        return "yaml" in content_type_name



import pytest
from transmute_core.function import TransmuteFunction
import transmute_core


@transmute_core.annotate({"return": int})
def raw_func():
    return 12345


@pytest.fixture
def func():
    return TransmuteFunction(raw_func)


def test_callable(func):
    """ test function is callable, and routes to the inner function. """
    assert func() == 12345


def test_return_type(func):
    """ test function is callable, and routes to the inner function. """
    assert func.return_type is int



import pytest
from transmute_core.contenttype_serializers import get_default_serializer_set
from transmute_core.schematics_serializer import SchematicsSerializer


@pytest.fixture
def serializer_set():
    return get_default_serializer_set()


@pytest.fixture
def serializer():
    return SchematicsSerializer()



from transmute_core.swagger import generate_swagger, get_swagger_static_root


def test_generate_swagger():
    swagger_static_root = "/statics/_swagger"
    swagger_json_url = "/swagger.json"
    body = generate_swagger(swagger_static_root, swagger_json_url)
    assert swagger_static_root in body
    assert swagger_json_url in body


def test_get_swagger_static_root():
    assert "static" in get_swagger_static_root()



from abc import ABCMeta, abstractmethod


class ContentTypeSerializer(object):
    """
    this class describes the interface that
    a serializer should follow to be allowed as a serializer.
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def to_type(data):
        """
        should return back a bytes (or string in python 2),
        representation of your object, to be used in e.g. response
        bodies.

        a ValueError should be returned in the case where
        the object cannote be serialized.
        """

    @abstractmethod
    def from_type(raw_bytes):
        """
        given a bytes object, should return a base python data
        structure that represents the object.

        a ValueError should be returned in the case where
        the object cannot be serialized.
        """

    @abstractmethod
    def can_handle(content_type_name):
        """
        given a content type, returns true if this serializer
        can convert bodies of the given type.
        """



from schematics.types import (
    BooleanType,
    IntType,
    FloatType,
    DecimalType,
    StringType
)
from schematics.exceptions import ConversionError
from .exceptions import SerializationException
from decimal import Decimal
from .compat import string_type

MODEL_MAP = {
    int: IntType(),
    bool: BooleanType(),
    float: FloatType(),
    Decimal: DecimalType(),
    string_type: StringType(),
}


class SchematicsSerializer(object):

    def __init__(self, builtin_models=None):
        builtin_models = builtin_models or MODEL_MAP
        self._models = dict(builtin_models)

    def load(self, model, value):
        try:
            if model in self._models:
                return self._models[model].to_native(value)
            return model(value)
        except ConversionError as e:
            raise SerializationException(str(e))

    def dump(self, model, value):
        try:
            if model in self._models:
                return self._models[model].to_primitive(value)
            return model.to_primitive(value)
        except ConversionError as e:
            raise SerializationException(str(e))



import pytest
from transmute_core.decorators import (
    PUT, POST, DELETE, annotate, TRANSMUTE_HTTP_METHOD_ATTRIBUTE
)


@pytest.mark.parametrize("decorator,method", [
    (PUT, "PUT"),
    (POST, "POST"),
    (DELETE, "DELETE")
])
def test_decorator(decorator, method):

    @decorator
    def test():
        pass

    assert getattr(test, TRANSMUTE_HTTP_METHOD_ATTRIBUTE) == set([method])


def test_joined_decorator():
    """ two decorators together should be merged """

    @POST
    @PUT
    def test():
        pass

    assert getattr(test, TRANSMUTE_HTTP_METHOD_ATTRIBUTE) == set(["POST", "PUT"])


def test_annotate():

    annotations = {"return": str, "arg": int}

    @annotate(annotations)
    def delete_test(arg):
        pass

    assert delete_test.__annotations__ == annotations



from schematics.models import Model
from schematics.types import StringType, IntType


def test_schematics_integration(serializer):

    class Card(Model):
        name = StringType()
        description = StringType()
        price = IntType()

    c = Card({
        "name": "foo", "description": "test",
        "price": 10}
    )

    assert serializer.dump(Card, c) == {
        "name": "foo",
        "description": "test",
        "price": 10
    }

    assert serializer.load(Card, {
        "name": "foo",
        "description": "test",
        "price": 10
    }) == c



from .function import TRANSMUTE_HTTP_METHOD_ATTRIBUTE


def PUT(f):
    """
    this labels a function as one that updates data.
    """
    _add_http_method_to_func("PUT", f)
    return f


def POST(f):
    """
    this labels a function as one that creates data.
    """
    _add_http_method_to_func("POST", f)
    return f


def DELETE(f):
    """
    this labels a function as one that deletes data.
    """
    _add_http_method_to_func("DELETE", f)
    return f


def _add_http_method_to_func(http_method, func):
    if not hasattr(func, TRANSMUTE_HTTP_METHOD_ATTRIBUTE):
        setattr(func, TRANSMUTE_HTTP_METHOD_ATTRIBUTE, set())
    getattr(func, TRANSMUTE_HTTP_METHOD_ATTRIBUTE).add(http_method)


def annotate(annotations):
    """
    in python2, native annotions on parameters do not exist:
    def foo(a : str, b: int) -> bool:
        ...

    this provides a way to provide attribute annotations:

    @annotate({"a": str, "b": int, "return": bool})
    def foo(a, b):
        ...
    """

    def decorate(func):
        func.__annotations__ = annotations
        return func

    return decorate



from transmute_core.exceptions import WebTransmuteException


class NoSerializerFound(WebTransmuteException):
    """
    raised when a serializer does not exist to handle the
    desired content type.
    """


class SerializerSet(object):
    """
    composes multiple serializers, delegating commands to one
    that can handle the desired content type.
    """

    def __init__(self, serializer_list):
        self._serializers = serializer_list

    def to_type(self, content_type, data):
        serializer = self.get_serializer_for_type(content_type)
        return serializer.to_type(data)

    def from_type(self, content_type, raw_bytes):
        serializer = self.get_serializer_for_type(content_type)
        return serializer.from_type(raw_bytes)

    def get_serializer_for_type(self, content_type):
        for serializer in self._serializers:
            if serializer.can_handle(content_type):
                return serializer
        raise NoSerializerFound(
            "unable to find serializer for " + content_type + "in: " +
            str(self._serializers)
        )



import os
import jinja2

CURDIR = os.path.dirname(__file__)


def generate_swagger(swagger_static_root, swagger_json_url):
    tmpl = get_template("swagger.html")
    return tmpl.render(
        swagger_root=swagger_static_root,
        swagger_json_url=swagger_json_url
    )


def get_swagger_static_root():
    return os.path.join(CURDIR, "static")

_template_cache = {}


def get_template(template_name):
    if template_name not in _template_cache:
        template_path = os.path.join(CURDIR, template_name)
        with open(template_path) as fh:
            _template_cache[template_name] = jinja2.Template(fh.read())
    return _template_cache[template_name]



import inspect
from .compat import getfullargspec
from .signature import get_signature
from swagger_schema import (
    Operation, Responses, Response, JsonSchemaObject
)


TRANSMUTE_HTTP_METHOD_ATTRIBUTE = "transmute_http_methods"


class TransmuteFunction(object):
    """
    VoodooFunctions are objects that wrap a method, allowing
    extensions to extract metadata for their own use (such as
    automatic documentation)
    """

    params = ["error_exceptions"]

    def __init__(self, func, error_exceptions=None):
        # arguments should be the arguments passed into
        # the function
        #
        # the return type should be specified. If nothing is
        # passed in, a NoneType is returned.
        # for a dynamic language like Python, it's not a huge deal:
        # transmute will still json serialize whatever object you pass it.
        # however, the return type is valuable for autodocumentation systems.
        self.argspec = getfullargspec(func)
        self.signature = get_signature(self.argspec)
        self.return_type = self.argspec.annotations.get("return")
        # description should be a description of the api
        # endpoint, for use in autodocumentation
        self.description = func.__doc__ or ""
        # error_exceptions represents the exceptions
        # that should be caught and return an API exception
        self.error_exceptions = error_exceptions
        # these are the http methods that should route to the given function.
        self.http_methods = getattr(
            func, TRANSMUTE_HTTP_METHOD_ATTRIBUTE, set(["GET"])
        )
        # produces represents the return types supported
        # by the final function
        self.produces = ["application/json"]
        self.raw_func = func
        self.swagger = self._get_swagger_operation()
        # this is to make discovery easier.
        # TODO: make sure this doesn't mess up GC, as it's
        # a cyclic reference.
        if inspect.ismethod(func):
            func.__func__.transmute_func = self
        else:
            func.transmute_func = self

    def __call__(self, *args, **kwargs):
        return self.raw_func(*args, **kwargs)

    def _get_swagger_operation(self):
        """ get the swagger_schema operation representation. """
        return Operation(
            summary=self.description,
            description=self.description,
            consumes=self.produces,
            produces=self.produces,
            responses=Responses({
                "200": Response(
                    description="success",
                    schema=JsonSchemaObject({
                        "properties": {
                            "success": {"type": "boolean"},
                            "result": {"type": "string"}
                        },
                        "required": ["success", "result"]
                    })
                ),
                "400": Response(
                    description="invalid input received",
                    schema=JsonSchemaObject({
                        "properties": {
                            "success": {"type": "boolean"},
                            "message": {"type": "string"}
                        },
                        "required": ["success", "message"]
                    })
                )
            })
        )



import pytest
from transmute_core.contenttype_serializers.serializer_set import NoSerializerFound


def test_default_serializer_json(serializer_set):
    frm, expected_to = {"foo": "bar"}, b'{"foo": "bar"}'
    assert serializer_set.to_type("application/json", frm) == expected_to
    assert serializer_set.from_type("application/json", expected_to) == frm


def test_default_serializer_yaml(serializer_set):
    frm, expected_to = {"foo": "bar"}, b'foo: bar\n'
    assert serializer_set.to_type("application/yaml", frm) == expected_to
    assert serializer_set.from_type("application/yaml", expected_to) == frm


def test_no_serializer_fonud_raises_exception(serializer_set):
    frm, _to = {"foo": "bar"}, b'foo: bar\n'
    with pytest.raises(NoSerializerFound):
        assert serializer_set.to_type("oogabooga", frm)

    with pytest.raises(NoSerializerFound):
        assert serializer_set.from_type("oogabooga", _to)



from transmute_core.compat import getfullargspec
from transmute_core import annotate
from transmute_core.signature import get_signature, NoDefault


def test_signature():

    @annotate({"x": int, "y": float, "width": int, "height": float})
    def make_square(x, y, width=None, height=12):
        pass

    argspec = getfullargspec(make_square)
    signature = get_signature(argspec)
    assert len(signature.args) == 2
    assert signature.args[0].name == "x"
    assert signature.args[0].default == NoDefault
    assert signature.args[0].type == int
    assert len(signature.kwargs) == 2
    assert signature.kwargs["width"].name == "width"
    assert signature.kwargs["width"].default is None
    assert signature.kwargs["width"].type == int


def test_signature_no_kwargs():

    @annotate({"x": int, "y": int, "return": int})
    def make_square(x, y):
        return x * y

    argspec = getfullargspec(make_square)
    signature = get_signature(argspec)
    assert len(signature.args) == 2
    assert signature.args[0].name == "x"
    assert signature.args[0].default == NoDefault
    assert signature.args[0].type == int
    assert signature.args[1].name == "y"
    assert signature.args[1].default == NoDefault
    assert signature.args[1].type == int
    assert len(signature.kwargs) == 0


def test_self_signature():

    def square(self, resource, multiplier=None):
        pass

    argspec = getfullargspec(square)
    signature = get_signature(argspec)

    assert len(signature.args) == 1
    assert len(signature.kwargs) == 1
    assert signature.kwargs["multiplier"].name == "multiplier"
